@@ -1,23 +1,33 @@
 from collections import OrderedDict
+from dataclasses import dataclass
 from typing import Dict, List, Optional
 
 import h5py
 import numpy as np
 from astromodels import TemplateModel, TemplateModelFactory
+from astromodels.utils.logging import silence_console_log
+from tqdm.auto import tqdm
 
-from .simulation_builder import Parameter, ParameterGrid
 from .utils.logging import setup_logger
 
 log = setup_logger(__name__)
 
 
+@dataclass(frozen=True)
+class ValueContainer:
+    params: np.ndarray
+    values: np.ndarray
+
+
 class Database:
     def __init__(
         self,
-        grid_points: Dict[str, np.ndarray],
+        grid_points: np.ndarray,
         parameter_names: List[str],
         energy_grid: np.ndarray,
-        values: Dict[str, np.ndarray],
+        run_time: np.ndarray,
+        values: np.ndarray,
+        meta_data: Optional[Dict[str, np.ndarray]] = None,
     ) -> None:
         """
         Databse of parameters and simulated values
@@ -45,15 +55,21 @@ class Database:
         # self._values: Dict[str, np.ndarray] = values
         # self._grid_points: Dict[str, np.ndarray] = grid_points
 
-        self._values: np.ndarray = np.empty(
-            (self._n_entries, len(self._energy_grid))
-        )
+        self._values: np.ndarray = values
 
-        self._grid_points: np.ndarray = np.empty(
-            (self._n_entries, self._n_parameters)
-        )
+        self._grid_points: np.ndarray = grid_points
 
-        self._extract_parameter_values(values, grid_points)
+        self._run_time: np.ndarray = run_time
+
+        self._parameter_ranges: Dict[str, np.ndarray] = OrderedDict()
+
+        for i, par in enumerate(self._parameter_names):
+
+            self._parameter_ranges[par] = np.sort(
+                np.unique(self._grid_points[:, i])
+            )
+
+        self._meta_data: Optional[Dict[str, np.ndarray]] = meta_data
 
     @property
     def n_entries(self) -> int:
@@ -67,34 +83,31 @@ class Database:
     def parameter_ranges(self) -> Dict[str, np.ndarray]:
         return self._parameter_ranges
 
-    def _extract_parameter_values(self, values, grid_points):
+    @property
+    def paramerter_names(self) -> List[str]:
+        return self._parameter_names
 
+    @property
+    def run_time(self) -> np.ndarray:
+        return self._run_time
+
+    @property
+    def energy_grid(self) -> np.ndarray:
+        return self._energy_grid
+
+    @property
+    def meta_data(self) -> Optional[Dict[str, np.ndarray]]:
+        return self._meta_data
+
+    def at(self, i: int) -> ValueContainer:
         """
-        extract the values and parameter ranges for
-        the tables
-
-        :param values:
-        :type values:
-        :param grid_points:
-        :type grid_points:
-        :returns:
-
+        return the parameters and values
+        at an index
         """
 
-        # extract the values
-
-        for i in range(self._n_entries):
-
-            self._values[i] = values[f"{i}"]
-            self._grid_points[i] = grid_points[f"{i}"]
-
-        self._parameter_ranges = OrderedDict()
-
-        for i, par in enumerate(self._parameter_names):
-
-            self._parameter_ranges[par] = np.sort(
-                np.unique(self._grid_points[:, i])
-            )
+        return ValueContainer(
+            params=self._grid_points[i, :], values=self._values[i, :]
+        )
 
     def _sub_selection(
         self,
@@ -151,18 +164,23 @@ class Database:
         return selection
 
     @classmethod
-    def from_file(cls, file_name: str) -> "Database":
-
+    def from_file(cls, file_name: str, output: int = 0) -> "Database":
+        """
+        open a database from a file.
+        The output argument specifies which value to collect.
+        For example, a simulation may save photon and electron
+        distributions from a solution.
+        """
         values = {}
         parameters = {}
 
         with h5py.File(file_name, "r") as f:
 
-            energy_grid = f['energy_grid'][()]
+            energy_grid = f['energy_grid'][f'energy_grid_{output}'][()]
 
             values_grp = f["values"]
 
-            parameters_grp = f["parameters"]
+            run_time = f["run_time"][()]
 
             par_name_grp = f["parameter_names"]
 
@@ -170,15 +188,36 @@ class Database:
                 par_name_grp.attrs[f"par{i}"]
                 for i in range(len(par_name_grp.attrs))
             ]
-            for key in values_grp.keys():
 
-                values[key] = values_grp[key][()]
-                parameters[key] = parameters_grp[key][()]
+            parameters = f['parameters'][()]
 
-        return cls(parameters, parameter_names, energy_grid, values)
+            values = values_grp[f'output_{output}'][()]
+
+            meta_data = None
+
+            if "meta" in list(f.keys()):
+
+                meta_data = {}
+
+                for k, v in f['meta'].items():
+
+                    meta_data[k] = v[()]
+
+        return cls(
+            grid_points=parameters,
+            parameter_names=parameter_names,
+            energy_grid=energy_grid,
+            values=values,
+            run_time=run_time,
+            meta_data=meta_data,
+        )
 
     def to_3ml(
-        self, name: str, desc: str, overwrite: bool = False, **kwargs
+        self,
+        name: str,
+        desc: str,
+        overwrite: bool = False,
+        **kwargs,
     ) -> TemplateModel:
 
         """
@@ -247,18 +286,108 @@ class Database:
 
             tmf.define_parameter_grid(k, sub_parameter_ranges[k])
 
-        for i in range(len(sub_values)):
+        with silence_console_log():
 
-            ### DO NOT SORT
+            for i in tqdm(range(len(sub_values))):
 
-            tmf.add_interpolation_data(
-                sub_values[i],
-                **{k: v for k, v in zip(self._parameter_names, sub_grid[i])},
-            )
+                ### DO NOT SORT
 
-        tmf.save_data(overwrite=True)
+                tmf.add_interpolation_data(
+                    sub_values[i],
+                    **{
+                        k: v for k, v in zip(self._parameter_names, sub_grid[i])
+                    },
+                )
+
+            tmf.save_data(overwrite=overwrite)
 
         return TemplateModel(name)
 
 
+def merge_databases(
+    *file_names: List[str], new_name: str = "merged_db.h5"
+) -> None:
+
+    """TODO describe function
+
+    :param new_name:
+    :type new_name: str
+    :returns:
+
+    """
+    n_entries = 0
+
+    for i, fname in enumerate(file_names):
+
+        with h5py.File(fname, "r") as f:
+
+            n_entries += f["parameters"].shape[0]
+
+            if i == 0:
+
+                par_name_grp = f["parameter_names"]
+
+                parameter_names = [
+                    par_name_grp.attrs[f"par{i}"]
+                    for i in range(len(par_name_grp.attrs))
+                ]
+
+                energy_grids = []
+
+                for _, v in f["energy_grid"].items():
+
+                    energy_grids.append(v[()])
+
+        n_outputs = len(energy_grids)
+
+        n_parameters = len(parameter_names)
+
+        parameters = np.zeros((n_entries, n_parameters))
+
+        values = []
+
+        for grid in energy_grids:
+
+            values.append(np.zeros((n_entries, len(grid))))
+
+        k = 0
+
+        # extract all the information
+
+        for i, fname in enumerate(file_names):
+
+            with h5py.File(fname, "r") as f:
+
+                n = f["parameters"].shape[0]
+
+                parameter_names[k:n, :] = f["parameters"][()]
+
+                for j, val in enumerate(values):
+
+                    val[k:n] = f["values"][f"output_{j}"][()]
+
+                k += n
+
+        # create the new file
+
+        with h5py.File(new_name, "w") as f:
+
+            f.create_dataset("parameters", data=parameters)
+
+            val_grp = f.create_group("values")
+            energy_grid_grp = f.create_group("energy_grid")
+
+            par_name_grp = f.create_group("parameters_names")
+            for i in range(n_parameters):
+
+                par_name_grp.attrs[f"par{i}"] = parameter_names[i]
+
+            for i in range(n_outputs):
+
+                val_grp.create_dataset(f"output_{i}", data=values[i])
+                energy_grid_grp.create_dataset(
+                    f"energy_grid_{i}", data=energy_grids[i]
+                )
+
+
 __all__ = ["Database"]
